from .common import (
    SDJWTCommon,
    DEFAULT_SIGNING_ALG,
    DIGEST_ALG_KEY,
    SD_DIGESTS_KEY,
    SD_LIST_PREFIX,
    KB_DIGEST_KEY,
)

from json import dumps, loads
from typing import Dict, List, Union, Callable

from jwcrypto.jwk import JWK
from jwcrypto.jws import JWS


class SDJWTVerifier(SDJWTCommon):
    _input_disclosures: List
    _hash_to_decoded_disclosure: Dict
    _hash_to_disclosure: Dict

    def __init__(
        self,
        sd_jwt_presentation: str,
        cb_get_issuer_key: Callable[[str, Dict], str],
        expected_aud: Union[str, None] = None,
        expected_nonce: Union[str, None] = None,
        serialization_format: str = "compact",
    ):
        super().__init__(serialization_format=serialization_format)

        self._parse_sd_jwt(sd_jwt_presentation)
        self._create_hash_mappings(self._input_disclosures)
        self._verify_sd_jwt(cb_get_issuer_key)

        # expected aud and nonce either need to be both set or both None
        if expected_aud or expected_nonce:
            if not (expected_aud and expected_nonce):
                raise ValueError(
                    "Either both expected_aud and expected_nonce must be provided or both must be None"
                )

            # Verify the SD-JWT-Release
            self._verify_key_binding_jwt(
                expected_aud,
                expected_nonce,
            )

    def get_verified_payload(self):
        return self._extract_sd_claims()

    def _verify_sd_jwt(
        self,
        cb_get_issuer_key,
        sign_alg: str = None,
    ):
        parsed_input_sd_jwt = JWS()
        parsed_input_sd_jwt.deserialize(self._unverified_input_sd_jwt)

        unverified_issuer = self._unverified_input_sd_jwt_payload.get("iss", None)
        unverified_header_parameters = parsed_input_sd_jwt.jose_header
        issuer_public_key = cb_get_issuer_key(unverified_issuer, unverified_header_parameters)
        parsed_input_sd_jwt.verify(issuer_public_key, alg=sign_alg)

        self._sd_jwt_payload = loads(parsed_input_sd_jwt.payload.decode("utf-8"))
        # TODO: Check exp/nbf/iat

        self._holder_public_key_payload = self._sd_jwt_payload.get("cnf", None)

    def _verify_key_binding_jwt(
        self,
        expected_aud: Union[str, None] = None,
        expected_nonce: Union[str, None] = None,
        sign_alg: Union[str, None] = None,
    ):

        # Deserialized the key binding JWT
        _alg = sign_alg or DEFAULT_SIGNING_ALG
        parsed_input_key_binding_jwt = JWS()
        parsed_input_key_binding_jwt.deserialize(self._unverified_input_key_binding_jwt)

        # Verify the key binding JWT using the holder public key
        if not self._holder_public_key_payload:
            raise ValueError("No holder public key in SD-JWT")

        holder_public_key_payload_jwk = self._holder_public_key_payload.get("jwk", None)
        if not holder_public_key_payload_jwk:
            raise ValueError(
                "The holder_public_key_payload is malformed. "
                "It doesn't contain the claim jwk: "
                f"{self._holder_public_key_payload}"
            )

        pubkey = JWK.from_json(dumps(holder_public_key_payload_jwk))

        parsed_input_key_binding_jwt.verify(pubkey, alg=_alg)

        # Check header typ
        key_binding_jwt_header = parsed_input_key_binding_jwt.jose_header

        if key_binding_jwt_header["typ"] != self.KB_JWT_TYP_HEADER:
            raise ValueError("Invalid header typ")

        # Check payload
        key_binding_jwt_payload = loads(parsed_input_key_binding_jwt.payload)

        if key_binding_jwt_payload["aud"] != expected_aud:
            raise ValueError("Invalid audience in KB-JWT")
        if key_binding_jwt_payload["nonce"] != expected_nonce:
            raise ValueError("Invalid nonce in KB-JWT")

        # Reassemble the SD-JWT in compact format and check digest
        if self._serialization_format == "compact":
            string_to_hash = self._combine(
                        self._unverified_input_sd_jwt,
                        *self._input_disclosures,
                        ""
            )
            expected_sd_jwt_presentation_hash = self._b64hash(string_to_hash.encode("ascii"))

            if key_binding_jwt_payload[KB_DIGEST_KEY] != expected_sd_jwt_presentation_hash:
                raise ValueError("Invalid digest in KB-JWT")

    def _extract_sd_claims(self):
        if DIGEST_ALG_KEY in self._sd_jwt_payload:
            if self._sd_jwt_payload[DIGEST_ALG_KEY] != self.HASH_ALG["name"]:
                # TODO: Support other hash algorithms
                raise ValueError("Invalid hash algorithm")

        self._duplicate_hash_check = []
        return self._unpack_disclosed_claims(self._sd_jwt_payload)

    def _unpack_disclosed_claims(self, sd_jwt_claims):
        # In a list, unpack each element individually
        if type(sd_jwt_claims) is list:
            output = []
            for element in sd_jwt_claims:
                if (
                    type(element) is dict
                    and len(element) == 1
                    and SD_LIST_PREFIX in element
                    and type(element[SD_LIST_PREFIX]) is str
                ):
                    digest_to_check = element[SD_LIST_PREFIX]
                    if digest_to_check in self._hash_to_decoded_disclosure:
                        _, value = self._hash_to_decoded_disclosure[digest_to_check]
                        output.append(self._unpack_disclosed_claims(value))
                else:
                    output.append(self._unpack_disclosed_claims(element))
            return output

        elif type(sd_jwt_claims) is dict:
            # First, try to figure out if there are any claims to be
            # disclosed in this dict. If so, replace them by their
            # disclosed values.

            pre_output = {
                k: self._unpack_disclosed_claims(v)
                for k, v in sd_jwt_claims.items()
                if k != SD_DIGESTS_KEY and k != DIGEST_ALG_KEY
            }

            for digest in sd_jwt_claims.get(SD_DIGESTS_KEY, []):
                if digest in self._duplicate_hash_check:
                    raise ValueError(f"Duplicate hash found in SD-JWT: {digest}")
                self._duplicate_hash_check.append(digest)

                if digest in self._hash_to_decoded_disclosure:
                    _, key, value = self._hash_to_decoded_disclosure[digest]
                    if key in pre_output:
                        raise ValueError(
                            f"Duplicate key found when unpacking disclosed claim: '{key}' in {pre_output}. This is not allowed."
                        )
                    unpacked_value = self._unpack_disclosed_claims(value)
                    pre_output[key] = unpacked_value

            # Now, go through the dict and unpack any nested dicts.

            return pre_output

        else:
            return sd_jwt_claims



#!/usr/bin/env python3
"""
This utility uses the SD-JWT library to update the static test case data in the
sd_jwt/test_cases directory. It is intended to be run after changes to the
library that affect the test cases.
"""


import argparse
import logging
import sys
from typing import Dict
from pathlib import Path

from sd_jwt import __version__
from sd_jwt.holder import SDJWTHolder
from sd_jwt.issuer import SDJWTIssuer
from sd_jwt.utils.demo_utils import get_jwk, load_yaml_settings
from sd_jwt.verifier import SDJWTVerifier

from sd_jwt.utils import formatting
from sd_jwt.utils.yaml_specification import (
    load_yaml_specification,
    remove_sdobj_wrappers,
)

logger = logging.getLogger("sd_jwt")

# Set logging to stdout
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


def generate_test_case_data(settings: Dict, testcase_path: Path, type: str):
    seed = settings["random_seed"]
    demo_keys = get_jwk(settings["key_settings"], True, seed)

    ### Load test case data
    testcase = load_yaml_specification(testcase_path)
    use_decoys = testcase.get("add_decoy_claims", False)
    serialization_format = testcase.get("serialization_format", "compact")
    include_default_claims = testcase.get("include_default_claims", True)
    extra_header_parameters = testcase.get("extra_header_parameters", {})

    claims = {}
    if include_default_claims:
        claims = {
            "iss": settings["identifiers"]["issuer"],
            "iat": settings["iat"],
            "exp": settings["exp"],
        }

    claims.update(testcase["user_claims"])

    ### Produce SD-JWT and SVC for selected example
    SDJWTIssuer.unsafe_randomness = True
    sdjwt_at_issuer = SDJWTIssuer(
        claims,
        demo_keys["issuer_key"],
        demo_keys["holder_key"] if testcase.get("key_binding", False) else None,
        add_decoy_claims=use_decoys,
        serialization_format=serialization_format,
        extra_header_parameters=extra_header_parameters,
    )

    ### Produce SD-JWT-R for selected example

    sdjwt_at_holder = SDJWTHolder(
        sdjwt_at_issuer.sd_jwt_issuance,
        serialization_format=serialization_format,
    )
    sdjwt_at_holder.create_presentation(
        testcase["holder_disclosed_claims"],
        settings["key_binding_nonce"]
        if testcase.get("key_binding", False)
        else None,
        settings["identifiers"]["verifier"]
        if testcase.get("key_binding", False)
        else None,
        demo_keys["holder_key"] if testcase.get("key_binding", False) else None,
    )

    ### Verify the SD-JWT using the SD-JWT-R

    # Define a function to check the issuer and retrieve the
    # matching public key
    def cb_get_issuer_key(issuer, header_parameters):
        # Do not use in production - this allows to use any issuer name for demo purposes
        if issuer == claims.get("iss", None):
            return demo_keys["issuer_public_key"]
        else:
            raise Exception(f"Unknown issuer: {issuer}")

    sdjwt_at_verifier = SDJWTVerifier(
        sdjwt_at_holder.sd_jwt_presentation,
        cb_get_issuer_key,
        settings["identifiers"]["verifier"]
        if testcase.get("key_binding", False)
        else None,
        settings["key_binding_nonce"]
        if testcase.get("key_binding", False)
        else None,
        serialization_format=serialization_format,
    )
    verified = sdjwt_at_verifier.get_verified_payload()

    # Write the test case data to the directory of the test case

    _artifacts = {
        "user_claims": (
            remove_sdobj_wrappers(testcase["user_claims"]),
            "User Claims",
            "json",
        ),
        "sd_jwt_payload": (
            sdjwt_at_issuer.sd_jwt_payload,
            "Payload of the SD-JWT",
            "json",
        ),
        "sd_jwt_jws_part": (
            sdjwt_at_issuer.serialized_sd_jwt,
            "Serialized SD-JWT",
            "txt" if serialization_format == "compact" else "json",
        ),
        "sd_jwt_issuance": (
            sdjwt_at_issuer.sd_jwt_issuance,
            "Combined SD-JWT and Disclosures",
            "txt" if serialization_format == "compact" else "json",
        ),
        "sd_jwt_presentation": (
            sdjwt_at_holder.sd_jwt_presentation,
            "Combined representation of SD-JWT and HS-Disclosures",
            "txt" if serialization_format == "compact" else "json",
        ),
        "verified_contents": (
            verified,
            "Verified released contents of the SD-JWT",
            "json",
        ),
    }

    if testcase.get("key_binding", False):
        _artifacts.update(
            {
                "kb_jwt_header": (
                    sdjwt_at_holder.key_binding_jwt_header
                    if testcase.get("key_binding")
                    else None,
                    "Header of the Holder Binding JWT",
                    "json",
                ),
                "kb_jwt_payload": (
                    sdjwt_at_holder.key_binding_jwt_payload
                    if testcase.get("key_binding")
                    else None,
                    "Payload of the Holder Binding JWT",
                    "json",
                ),
                "kb_jwt_serialized": (
                    sdjwt_at_holder.serialized_key_binding_jwt,
                    "Serialized Holder Binding JWT",
                    "txt",
                ),
            }
        )

    # When type is example, add info about disclosures
    if type == "example":
        _artifacts["disclosures"] = (
            formatting.markdown_disclosures(
                sdjwt_at_issuer.ii_disclosures,
            ),
            "Payloads of the II-Disclosures",
            "md",
        )

    # When decoys were used, list those as well (here as a json array)
    if use_decoys:
        if type == "example":
            _artifacts.update(
                {
                    "decoy_digests": (
                        formatting.markdown_decoy_digests(
                            sdjwt_at_issuer.decoy_digests
                        ),
                        "Decoy Claims",
                        "md",
                    )
                }
            )
        else:
            _artifacts.update(
                {
                    "decoy_digests": (
                        sdjwt_at_issuer.decoy_digests,
                        "Decoy Claims",
                        "json",
                    )
                }
            )

    output_dir = testcase_path.parent

    logger.info(f"Writing test case data to '{output_dir}'.")

    if not output_dir.exists():
        sys.exit(f"Output directory '{output_dir}' does not exist.")

    formatter = (
        formatting.format_for_example
        if type == "example"
        else formatting.format_for_testcase
    )

    for key, data_item in _artifacts.items():
        if data_item is None:
            continue

        logger.info(f"Writing {key} to '{output_dir / key}'.")

        data, _, ftype = data_item

        with open(output_dir / f"{key}.{ftype}", "w") as f:
            f.write(formatter(data, ftype))


# For all *.yml files in subdirectories of the working directory, run the test case generation
def run():
    # This tool must be called with either "testcase" or "example" as the first argument in order
    # to specify which type of output to generate.

    parser = argparse.ArgumentParser(
        description=(
            "Generate test cases or examples for SD-JWT library. "
            "Test case data is suitable for use in other SD-JWT libraries. "
            "Examples are formatted in a markdown-friendly way (e.g., line breaks, "
            "markdown formatting) for direct inclusion into the specification text."
        )
    )

    # Type is a positional argument, either testcase or example
    parser.add_argument(
        "type",
        choices=["testcase", "example"],
        help="Whether to generate test cases or examples.",
    )

    # Optional: One or more names of directories containing test cases to generate
    parser.add_argument(
        "directories",
        nargs="*",
        help=(
            "One or more names of directories containing test cases to generate. "
            "If no directories are specified, all directories containing a file "
            "named 'specification.yml' respectively are processed."
        ),
    )
    args = parser.parse_args()

    basedir = Path.cwd()
    settings_file = basedir / "settings.yml"

    if not settings_file.exists():
        sys.exit(f"Settings file '{settings_file}' does not exist.")

    if args.directories:
        glob = [basedir / d / "specification.yml" for d in args.directories]
    else:
        glob = basedir.glob("*/specification.yml")

    # load keys and other information from test_settings.yml
    settings = load_yaml_settings(settings_file)

    for case_path in glob:
        logger.info(f"Generating data for '{case_path}'")
        generate_test_case_data(settings, case_path, args.type)


if __name__ == "__main__":
    run()



import base64
import json
import logging
import random
import yaml
import sys

from jwcrypto.jwk import JWK
from typing import Union

logger = logging.getLogger("sd_jwt")


def load_yaml_settings(file):
    with open(file, "r") as f:
        settings = yaml.safe_load(f)

    for property in ("identifiers", "key_settings"):
        if property not in settings:
            sys.exit(f"Settings file must define '{property}'.")

    return settings


def print_repr(values: Union[str, list], nlines=2):
    value = "\n".join(values) if isinstance(values, (list, tuple)) else values
    _nlines = "\n" * nlines if nlines else ""
    print(value, end=_nlines)


def print_decoded_repr(value: str, nlines=2):
    seq = []
    for i in value.split("."):
        try:
            padded = f"{i}{'=' * divmod(len(i),4)[1]}"
            seq.append(f"{base64.urlsafe_b64decode(padded).decode()}")
        except Exception as e:
            logging.debug(f"{e} - for value: {i}")
            seq.append(i)
    _nlines = "\n" * nlines if nlines else ""
    print("\n.\n".join(seq), end=_nlines)


def get_jwk(jwk_kwargs: dict = {}, no_randomness: bool = False, random_seed: int = 0):
    """
    jwk_kwargs = {
        issuer_key:dict : {},
        holder_key:dict : {},
        key_size: int : 0,
        kty: str : "RSA"
    }

    returns static or random JWK
    """
    if no_randomness:
        random.seed(random_seed)
        issuer_key = JWK.from_json(json.dumps(jwk_kwargs["issuer_key"]))
        holder_key = JWK.from_json(json.dumps(jwk_kwargs["holder_key"]))
        logger.warning("Using fixed randomness for demo purposes")
    else:
        _kwargs = {"key_size": jwk_kwargs["key_size"], "kty": jwk_kwargs["kty"]}
        issuer_key = JWK.generate(**_kwargs)
        holder_key = JWK.generate(**_kwargs)

    issuer_public_key = JWK.from_json(issuer_key.export_public())
    return dict(
        issuer_key=issuer_key,
        holder_key=holder_key,
        issuer_public_key=issuer_public_key,
    )



from dataclasses import dataclass
from json import dumps
from typing import Optional


@dataclass
class SDJWTDisclosure:
    """This class represents a disclosure of a claim."""

    issuer: any
    key: Optional[str]  # only for object keys
    value: any

    def __post_init__(self):
        self._hash()

    def _hash(self):
        salt = self.issuer._generate_salt()
        if self.key is None:
            data = [salt, self.value]
        else:
            data = [salt, self.key, self.value]

        self._json = dumps(data).encode("utf-8")

        self._raw_b64 = self.issuer._base64url_encode(self._json)
        self._hash = self.issuer._b64hash(self._raw_b64.encode("ascii"))

    @property
    def hash(self):
        return self._hash

    @property
    def b64(self):
        return self._raw_b64

    @property
    def json(self):
        return self._json.decode("utf-8")



import json
import logging
from textwrap import fill, wrap
from pathlib import Path

from sd_jwt.common import SDObj

logger = logging.getLogger("sd_jwt")

OUTPUT_INDENT = 2
EXAMPLE_MAX_WIDTH = 68
EXAMPLE_SHORT_WIDTH = 60
OUTPUT_ENSURE_ASCII = False

#######################################################################
# Helper functions to format examples
#######################################################################


def textwrap_json(data, width=EXAMPLE_MAX_WIDTH):
    text = json.dumps(data, indent=OUTPUT_INDENT, ensure_ascii=OUTPUT_ENSURE_ASCII)
    output = []
    for line in text.splitlines():
        if len(line) <= width:
            output.append(line)
        else:
            # Check if line is of the form "key": "value"
            if not line.strip().startswith('"'):
                logger.warning("unexpected line " + line)
                output.append(line)
                continue
            # Determine number of spaces before the value
            ##spaces = line.index(":") + 2
            spaces = line.index('"') + OUTPUT_INDENT
            # Wrap the value
            wrapped = wrap(
                line[spaces:],
                width=width - spaces,
                break_on_hyphens=False,
                replace_whitespace=False,
            )
            # Add the wrapped value to the output
            output.append(line[:spaces] + wrapped[0])
            for line in wrapped[1:]:
                output.append(" " * spaces + line)
    output = "\n".join(text for text in output)

    return output


def textwrap_text(text, width=EXAMPLE_MAX_WIDTH):
    return fill(
        text,
        width=width,
        break_on_hyphens=False,
    )


def multiline_code(text):
    # Add a ` character to each start and end of a line and a backslash after each line
    return "\\\n".join(f"`{line}`" for line in text.splitlines())


def markdown_disclosures(disclosures):
    markdown = ""
    for d in disclosures:
        if d.key is None:
            markdown += f"__Array Entry__:\n\n"
        else:
            markdown += f"__Claim `{d.key}`__:\n\n"

        markdown += (
            f" * SHA-256 Hash: `{d.hash}`\n"
            f" * Disclosure:\\\n"
            f"{multiline_code(textwrap_text(d.b64, EXAMPLE_SHORT_WIDTH))}\n"
            f" * Contents:\n"
            f"{multiline_code(textwrap_text(d.json, EXAMPLE_SHORT_WIDTH))}\n\n\n"
        )

    return markdown.strip()


def markdown_decoy_digests(decoy_digests):
    # create a list of decoy digests in markdown format
    return "\n".join(f" * `{digest}`" for digest in decoy_digests)


def format_for_testcase(data, ftype):
    if ftype == "json":
        return json.dumps(data, indent=OUTPUT_INDENT, ensure_ascii=OUTPUT_ENSURE_ASCII)
    else:
        return data


def format_for_example(data, ftype):
    if ftype == "json":
        if isinstance(data, str):
            data = json.loads(data)
        return textwrap_json(data)
    elif ftype == "txt":
        return textwrap_text(data)
    else:
        return data



#!/usr/bin/env python3
import argparse
import logging
import pathlib
import random
import sys
import datetime


from sd_jwt import __version__
from sd_jwt.utils.demo_utils import (
    get_jwk,
    print_decoded_repr,
    load_yaml_settings,
)
from sd_jwt.holder import SDJWTHolder
from sd_jwt.issuer import SDJWTIssuer
from sd_jwt.verifier import SDJWTVerifier

from sd_jwt.utils.formatting import (
    textwrap_json,
    textwrap_text,
    multiline_code,
    EXAMPLE_SHORT_WIDTH,
)
from sd_jwt.utils.yaml_specification import load_yaml_specification

logger = logging.getLogger("sd_jwt")


# Generate a 16-bit random number
def generate_nonce():
    return bytes(random.getrandbits(8) for _ in range(16)).hex()


DEFAULT_EXP_MINS = 15

def run():
    parser = argparse.ArgumentParser(
        description=f"{__file__} demo.",
        epilog=f"{__file__}",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "example",
        help=(
            "Yaml file containing the SD-JWT demo to process. See examples/simple.yml for an example."
        ),
        type=pathlib.Path,
    )
    parser.add_argument(
        "-d",
        "--debug",
        required=False,
        choices=("CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"),
        default="INFO",
        help="Debug level, see python logging; defaults to INFO if omitted",
    )
    parser.add_argument(
        "-nr",
        "--no-randomness",
        required=False,
        action="store_true",
        default=False,
        help=(
            "For the purpose of generating static examples for the spec, this command line "
            "switch disables randomness. Using this in production is highly insecure!"
        ),
    )
    parser.add_argument(
        "--nonce",
        required=False,
        type=str,
        default=generate_nonce(),
        help=("given example of a nonce: 'XZOUco1u_gEPknxS78sWWg'"),
    )
    parser.add_argument(
        "--iat", required=False, type=int, help=("issued at, UTC Timestamp")
    )
    parser.add_argument(
        "--exp", required=False, type=int, help=("expire at, UTC Timestamp")
    )
    parser.add_argument(
        "--settings-path",
        required=False,
        type=str,
        help=("Path to YAML file containing keys and other settings for the demo."),
        default="utils/demo_settings.yml",
    )
    parser.add_argument(
        "--indent",
        required=False,
        type=int,
        default=4,
        help=("json output indentation level"),
    )
    # new option to put examples into a directory
    parser.add_argument(
        "--output-dir",
        required=False,
        type=pathlib.Path,
        help=(
            "path/to/directory - Write all the examples into separate files in this directory"
        ),
    )
    parser.add_argument(
        "-v",
        "--version",
        required=False,
        action="store_true",
        help="Print version and exit",
    )

    _args = parser.parse_args()
    logger.setLevel(_args.debug)

    if _args.version:
        sys.exit(f"{__version__}")

    ### Load settings
    settings = load_yaml_settings(_args.settings_path)

    # If "no randomness" is requested, we hash the file name of the example
    # file to use it as the random seed. This ensures that the same example
    # file always generates the same output, but the output between
    # different example files is different.
    if _args.no_randomness:
        import hashlib

        hash_object = hashlib.sha256(_args.example.read_bytes())
        # Extract the hash as integer
        seed = int(hash_object.hexdigest(), 16)
    else:
        seed = None

    demo_keys = get_jwk(settings["key_settings"], _args.no_randomness, seed)

    ### Load example file

    example_identifer = _args.example.stem
    example = load_yaml_specification(_args.example)
    use_decoys = example.get("add_decoy_claims", False)

    ### Add default claims if necessary
    iat = _args.iat or int(datetime.datetime.utcnow().timestamp())
    exp = _args.exp or iat + (DEFAULT_EXP_MINS * 60)
    claims = {
        "iss": settings.ISSUER,
        "iat": iat,
        "exp": exp,
    }

    claims.update(example["user_claims"])

    ### Produce SD-JWT and SVC for selected example
    SDJWTIssuer.unsafe_randomness = _args.no_randomness
    sdjwt_at_issuer = SDJWTIssuer(
        claims,
        demo_keys["issuer_key"],
        demo_keys["holder_key"] if example.get("key_binding", False) else None,
        add_decoy_claims=use_decoys,
    )

    ### Produce SD-JWT-R for selected example

    # Note: The only input from the issuer is the combined SD-JWT and SVC!

    sdjwt_at_holder = SDJWTHolder(sdjwt_at_issuer.sd_jwt_issuance)
    sdjwt_at_holder.create_presentation(
        example["holder_disclosed_claims"],
        _args.nonce if example.get("key_binding", False) else None,
        settings.VERIFIER if example.get("key_binding", False) else None,
        demo_keys["holder_key"] if example.get("key_binding", False) else None,
    )

    ### Verify the SD-JWT using the SD-JWT-R


    # Define a function to check the issuer and retrieve the
    # matching public key
    def cb_get_issuer_key(issuer):
        # Do not use in production - this allows to use any issuer name for demo purposes
        if issuer == claims["iss"]:
            return demo_keys["issuer_public_key"]
        else:
            raise Exception(f"Unknown issuer: {issuer}")


    # Note: The only input from the holder is the combined presentation!
    sdjwt_at_verifier = SDJWTVerifier(
        sdjwt_at_holder.sd_jwt_presentation,
        cb_get_issuer_key,
        settings.VERIFIER if example.get("key_binding", False) else None,
        _args.nonce if example.get("key_binding", False) else None,
    )
    verified = sdjwt_at_verifier.get_verified_payload()

    ### Done - now output everything to CLI (unless --replace-examples-in was used)

    iid_payload = ""
    for hash in sdjwt_at_holder._hash_to_decoded_disclosure:
        salt, claim_name, claim_value = sdjwt_at_holder._hash_to_decoded_disclosure[hash]
        b64 = sdjwt_at_holder._hash_to_disclosure[hash]
        encoded_json = sdjwt_at_holder._base64url_decode(b64).decode("utf-8")

        iid_payload += (
            f"__Claim `{claim_name}`:__\n\n"
            f" * SHA-256 Hash: `{hash}`\n"
            f" * Disclosure:\\\n"
            f"{multiline_code(textwrap_text(b64, EXAMPLE_SHORT_WIDTH))}\n"
            f" * Contents:\n"
            f"{multiline_code(textwrap_text(encoded_json, EXAMPLE_SHORT_WIDTH))}\n\n\n"
        )

    iid_payload = iid_payload.strip()

    _artifacts = {
        "user_claims": (example["user_claims"], "User Claims", "json"),
        "sd_jwt_payload": (sdjwt_at_issuer.sd_jwt_payload, "Payload of the SD-JWT", "json"),
        "sd_jwt_jws_part": (
            sdjwt_at_issuer.serialized_sd_jwt,
            "Serialized SD-JWT",
            "txt",
        ),
        "disclosures": (iid_payload, "Payloads of the II-Disclosures", "md"),
        "sd_jwt_issuance": (
            sdjwt_at_issuer.sd_jwt_issuance,
            "Combined SD-JWT and Disclosures",
            "txt",
        ),
        "kb_jwt_payload": (
            sdjwt_at_holder.key_binding_jwt_payload
            if example.get("key_binding")
            else None,
            "Payload of the Holder Binding JWT",
            "json",
        ),
        "kb_jwt_serialized": (
            sdjwt_at_holder.serialized_key_binding_jwt,
            "Serialized Holder Binding JWT",
            "txt",
        ),
        "sd_jwt_presentation": (
            sdjwt_at_holder.sd_jwt_presentation,
            "Combined representation of SD-JWT and HS-Disclosures",
            "txt",
        ),
        "verified_contents": (verified, "Verified released contents of the SD-JWT", "json"),
    }

    # When decoys were used, list those as well
    if use_decoys:
        # create a list of decoy digests in markdown format
        decoy_digests = ""

        for digest in sdjwt_at_issuer.decoy_digests:
            decoy_digests += f" * `{digest}`\n"

        _artifacts["decoy_digests"] = (
            decoy_digests,
            "Decoy Claims",
            "md",
        )


    if _args.output_dir:
        logger.info(
            f"Writing all the examples into separate files in '{_args.output_dir}'."
        )

        output_dir = _args.output_dir / example_identifer

        if not output_dir.exists():
            output_dir.mkdir(parents=True)

        for key, (data, _, ftype) in _artifacts.items():
            if data is None:
                continue

            if ftype == "json":
                out = textwrap_json(data)
            elif ftype == "txt":
                out = textwrap_text(data)
            else:
                out = data

            with open(output_dir / f"{key}.{ftype}", "w") as f:
                f.write(out)

    else:
        for key, (data, description, ftype) in _artifacts.items():
            print(f"{description} ({key}):")
            if ftype == "json":
                out = textwrap_json(data)
            elif ftype == "txt":
                out = textwrap_text(data)
            else:
                out = data

            print(out)

            # Small hack to display some values in decoded form
            if key.startswith("serialized_"):
                print(" - decodes to - ")
                print_decoded_repr(data)

        sys.exit(0)

if __name__ == "__main__":
    run()


import logging 

from .common import (
    SDJWTCommon, 
    DEFAULT_SIGNING_ALG, 
    SD_DIGESTS_KEY, 
    SD_LIST_PREFIX, 
    KB_DIGEST_KEY
)
from json import dumps, loads
from time import time
from typing import Dict, List, Optional
from itertools import zip_longest

from jwcrypto.jws import JWS

logger = logging.getLogger("sd_jwt")


class SDJWTHolder(SDJWTCommon):
    hs_disclosures: List
    key_binding_jwt_header: Dict
    key_binding_jwt_payload: Dict
    key_binding_jwt: JWS
    serialized_key_binding_jwt: str = ""
    sd_jwt_presentation: str

    _input_disclosures: List
    _hash_to_decoded_disclosure: Dict
    _hash_to_disclosure: Dict

    def __init__(self, sd_jwt_issuance: str, serialization_format: str = "compact"):
        super().__init__(serialization_format=serialization_format)

        self._parse_sd_jwt(sd_jwt_issuance)

        # TODO: This holder does not verify the SD-JWT yet - this
        # is not strictly needed, but it would be nice to have.
        self.serialized_sd_jwt = self._unverified_input_sd_jwt
        self.sd_jwt_payload = self._unverified_input_sd_jwt_payload
        if self._serialization_format == "json":
            self.sd_jwt_parsed = self._unverified_input_sd_jwt_parsed

        self._create_hash_mappings(self._input_disclosures)

    def create_presentation(
        self, claims_to_disclose, nonce=None, aud=None, holder_key=None, sign_alg=None
    ):
        # Select the disclosures
        self.hs_disclosures = []
        self._select_disclosures(self.sd_jwt_payload, claims_to_disclose)

        # Optional: Create a key binding JWT
        if nonce and aud and holder_key:
            # Temporarily create the combined presentation in order to create the hash over it
            string_to_hash = self._combine(
                    self.serialized_sd_jwt,
                    *self.hs_disclosures,
                    ""
                )
            sd_jwt_presentation_hash = self._b64hash(string_to_hash.encode("ascii"))
            self._create_key_binding_jwt(nonce, aud, sd_jwt_presentation_hash, holder_key, sign_alg)


        # Create the combined presentation
        if self._serialization_format == "compact":
            # Note: If the key binding JWT is not created, then the
            # last element is empty, matching the spec.
            self.sd_jwt_presentation = self._combine(
                self.serialized_sd_jwt,
                *self.hs_disclosures,
                self.serialized_key_binding_jwt,
            )
        else:
            # In this case, take the parsed JSON serialized SD-JWT and
            # only filter the disclosures in the header. Add the holder
            # binding JWT to the header if it was created.
            self.sd_jwt_parsed[self.JWS_KEY_DISCLOSURES] = self.hs_disclosures
            if self.serialized_key_binding_jwt:
                self.sd_jwt_parsed[
                    self.JWS_KEY_KB_JWT
                ] = self.serialized_key_binding_jwt
            self.sd_jwt_presentation = dumps(self.sd_jwt_parsed)

    def _select_disclosures(self, sd_jwt_claims, claims_to_disclose):
        # Recursively process the claims in sd_jwt_claims. In each
        # object found therein, look at the SD_DIGESTS_KEY. If it
        # contains hash digests for claims that should be disclosed,
        # then add the corresponding disclosures to the claims_to_disclose.

        if type(sd_jwt_claims) is list:
            return self._select_disclosures_list(sd_jwt_claims, claims_to_disclose)
        elif type(sd_jwt_claims) is dict:
            return self._select_disclosures_dict(sd_jwt_claims, claims_to_disclose)
        else:
            pass

    def _select_disclosures_list(self, sd_jwt_claims, claims_to_disclose):
        if claims_to_disclose is None:
            return []
        if claims_to_disclose is True:
            claims_to_disclose = []
        if not type(claims_to_disclose) is list:
            raise ValueError(
                f"To disclose array elements, an array must be provided as disclosure information.\n"
                f"Found {claims_to_disclose} instead.\n"
                f"Check disclosure information for array: {sd_jwt_claims}"
            )

        for pos, (claims_to_disclose_element, element) in enumerate(
            zip_longest(claims_to_disclose, sd_jwt_claims, fillvalue=None)
        ):
            if (
                isinstance(element, dict)
                and len(element) == 1
                and SD_LIST_PREFIX in element
                and type(element[SD_LIST_PREFIX]) is str
            ):
                digest_to_check = element[SD_LIST_PREFIX]
                if digest_to_check not in self._hash_to_decoded_disclosure:
                    # fake digest
                    continue

                # Determine type of disclosure
                _, disclosure_value = self._hash_to_decoded_disclosure[digest_to_check]

                # Disclose the claim only if in claims_to_disclose (assumed to be an array)
                # there is an element with the current index and it is not None or False
                if claims_to_disclose_element in (
                    False,
                    None,
                ):
                    continue

                self.hs_disclosures.append(self._hash_to_disclosure[digest_to_check])
                if isinstance(disclosure_value, dict):
                    if claims_to_disclose_element is True:
                        # Tolerate a "True" for a disclosure of an object
                        claims_to_disclose_element = {}
                    if not isinstance(claims_to_disclose_element, dict):
                        raise ValueError(
                            f"To disclose object elements in arrays, provide an object (can be empty).\n"
                            f"Found {claims_to_disclose_element} instead.\n"
                            f"Problem at position {pos} of {claims_to_disclose}.\n"
                            f"Check disclosure information for object: {sd_jwt_claims}"
                        )
                    self._select_disclosures(
                        disclosure_value, claims_to_disclose_element
                    )
                elif isinstance(disclosure_value, list):
                    if claims_to_disclose_element is True:
                        # Tolerate a "True" for a disclosure of an array
                        claims_to_disclose_element = []
                    if not isinstance(claims_to_disclose_element, list):
                        raise ValueError(
                            f"To disclose array elements nested in arrays, provide an array (can be empty).\n"
                            f"Found {claims_to_disclose_element} instead.\n"
                            f"Problem at position {pos} of {claims_to_disclose}.\n"
                            f"Check disclosure information for array: {sd_jwt_claims}"
                        )

                    self._select_disclosures(
                        disclosure_value, claims_to_disclose_element
                    )

            else:
                self._select_disclosures(element, claims_to_disclose_element)

    def _select_disclosures_dict(self, sd_jwt_claims, claims_to_disclose):
        if claims_to_disclose is None:
            return {}
        if claims_to_disclose is True:
            # Tolerate a "True" for a disclosure of an object
            claims_to_disclose = {}
        if not isinstance(claims_to_disclose, dict):
            raise ValueError(
                f"To disclose object elements, an object must be provided as disclosure information.\n"
                f"Found {claims_to_disclose} (type {type(claims_to_disclose)}) instead.\n"
                f"Check disclosure information for object: {sd_jwt_claims}"
            )
        for key, value in sd_jwt_claims.items():
            if key == SD_DIGESTS_KEY:
                for digest_to_check in value:
                    if digest_to_check not in self._hash_to_decoded_disclosure:
                        # fake digest
                        continue
                    _, key, value = self._hash_to_decoded_disclosure[digest_to_check]

                    try:
                        logger.debug(
                            f"In _select_disclosures_dict: {key}, {value}, {claims_to_disclose}"
                        )
                        if key in claims_to_disclose and claims_to_disclose[key]:
                            logger.debug(f"Adding disclosure for {digest_to_check}")
                            self.hs_disclosures.append(
                                self._hash_to_disclosure[digest_to_check]
                            )
                        else:
                            logger.debug(
                                f"Not adding disclosure for {digest_to_check}, {key} (type {type(key)}) not in {claims_to_disclose}"
                            )
                    except TypeError:
                        # claims_to_disclose is not a dict
                        raise TypeError(
                            f"claims_to_disclose does not contain a dict where a dict was expected (found {claims_to_disclose} instead)\n"
                            f"Check claims_to_disclose for key: {key}, value: {value}"
                        ) from None

                    self._select_disclosures(value, claims_to_disclose.get(key, None))
            else:
                self._select_disclosures(value, claims_to_disclose.get(key, None))

    def _create_key_binding_jwt(
        self, nonce, aud, presentation_hash, holder_key, sign_alg: Optional[str] = None
    ):
        _alg = sign_alg or DEFAULT_SIGNING_ALG

        self.key_binding_jwt_header = {
            "alg": _alg,
            "typ": self.KB_JWT_TYP_HEADER,
        }

        self.key_binding_jwt_payload = {
            "nonce": nonce,
            "aud": aud,
            "iat": int(time()),
            KB_DIGEST_KEY: presentation_hash,
        }

        # Sign the SD-JWT-Release using the holder's key
        self.key_binding_jwt = JWS(
            payload=dumps(self.key_binding_jwt_payload),
        )

        self.key_binding_jwt.add_signature(
            holder_key,
            alg=_alg,
            protected=dumps(self.key_binding_jwt_header),
        )
        self.serialized_key_binding_jwt = self.key_binding_jwt.serialize(compact=True)



import random
from json import loads, dumps
from typing import Dict, List

from jwcrypto.jws import JWS

from .common import (
    DEFAULT_SIGNING_ALG,
    DIGEST_ALG_KEY,
    SD_DIGESTS_KEY,
    SD_LIST_PREFIX,
    SDJWTCommon,
    SDObj,
)
from .disclosure import SDJWTDisclosure


class SDJWTIssuer(SDJWTCommon):
    DECOY_MIN_ELEMENTS = 2
    DECOY_MAX_ELEMENTS = 5

    sd_jwt_payload: Dict
    sd_jwt: JWS
    serialized_sd_jwt: str

    ii_disclosures: List
    sd_jwt_issuance: str

    decoy_digests: List

    def __init__(
        self,
        user_claims: Dict,
        issuer_key,
        holder_key=None,
        sign_alg=None,
        add_decoy_claims: bool = False,
        serialization_format: str = "compact",
        extra_header_parameters: dict = {},
    ):
        super().__init__(serialization_format=serialization_format)

        self._user_claims = user_claims
        self._issuer_key = issuer_key
        self._holder_key = holder_key
        self._sign_alg = sign_alg or DEFAULT_SIGNING_ALG
        self._add_decoy_claims = add_decoy_claims
        self._extra_header_parameters = extra_header_parameters

        self.ii_disclosures = []
        self.decoy_digests = []

        self._check_for_sd_claim(self._user_claims)
        self._assemble_sd_jwt_payload()
        self._create_signed_jws()
        self._create_combined()

    def _assemble_sd_jwt_payload(self):
        # Create the JWS payload
        self.sd_jwt_payload = self._create_sd_claims(self._user_claims)
        self.sd_jwt_payload.update(
            {
                DIGEST_ALG_KEY: self.HASH_ALG["name"],
            }
        )
        if self._holder_key:
            self.sd_jwt_payload["cnf"] = {
                "jwk": self._holder_key.export_public(as_dict=True)
            }

    def _create_decoy_claim_entry(self) -> str:
        digest = self._b64hash(self._generate_salt().encode("ascii"))
        self.decoy_digests.append(digest)
        return digest

    def _create_sd_claims(self, user_claims):
        # This function can be called recursively.
        #
        # If the user claims are a list, apply this function
        # to each item in the list.
        if isinstance(user_claims, list):
            return self._create_sd_claims_list(user_claims)

        # If the user claims are a dictionary, apply this function
        # to each key/value pair in the dictionary.
        elif isinstance(user_claims, dict):
            return self._create_sd_claims_object(user_claims)

        # For other types, assume that the value can be disclosed.
        elif isinstance(user_claims, SDObj):
            raise ValueError(
                f"SDObj found in illegal place.\nThe claim value '{user_claims}' should not be wrapped by SDObj."
            )
        return user_claims

    def _create_sd_claims_list(self, user_claims: List):
        # Walk through all elements in the list.
        # If an element is marked as SD, then create a proper disclosure for it.
        # Otherwise, just return the element.

        output_user_claims = []
        for claim in user_claims:
            if isinstance(claim, SDObj):
                subtree_from_here = self._create_sd_claims(claim.value)
                # Create a new disclosure
                disclosure = SDJWTDisclosure(
                    self,
                    key=None,
                    value=subtree_from_here,
                )

                # Add to ii_disclosures
                self.ii_disclosures.append(disclosure)

                # Assemble all hash digests in the disclosures list.
                output_user_claims.append({SD_LIST_PREFIX: disclosure.hash})
            else:
                subtree_from_here = self._create_sd_claims(claim)
                output_user_claims.append(subtree_from_here)

        return output_user_claims

    def _create_sd_claims_object(self, user_claims: Dict):
        sd_claims = {SD_DIGESTS_KEY: []}
        for key, value in user_claims.items():
            subtree_from_here = self._create_sd_claims(value)
            if isinstance(key, SDObj):
                # Create a new disclosure
                disclosure = SDJWTDisclosure(
                    self,
                    key=key.value,
                    value=subtree_from_here,
                )

                # Add to ii_disclosures
                self.ii_disclosures.append(disclosure)

                # Assemble all hash digests in the disclosures list.
                sd_claims[SD_DIGESTS_KEY].append(disclosure.hash)
            else:
                sd_claims[key] = subtree_from_here

        # Add decoy claims if requested
        if self._add_decoy_claims:
            for _ in range(
                random.randint(self.DECOY_MIN_ELEMENTS, self.DECOY_MAX_ELEMENTS)
            ):
                sd_claims[SD_DIGESTS_KEY].append(self._create_decoy_claim_entry())

        # Delete the SD_DIGESTS_KEY if it is empty
        if len(sd_claims[SD_DIGESTS_KEY]) == 0:
            del sd_claims[SD_DIGESTS_KEY]
        else:
            # Sort the hash digests otherwise
            sd_claims[SD_DIGESTS_KEY].sort()

        return sd_claims

    def _create_signed_jws(self):
        """
        Create the SD-JWT.

        If serialization_format is "compact", then the SD-JWT is a JWT (JWS in compact serialization).
        If serialization_format is "json", then the SD-JWT is a JWS in JSON serialization. The disclosures in this case
        will be added in a separate "disclosures" property of the JSON.
        """

        self.sd_jwt = JWS(payload=dumps(self.sd_jwt_payload))

        # Assemble protected headers starting with default
        _protected_headers = {
            "alg": self._sign_alg,
            "typ": self.SD_JWT_HEADER
        }
        # override if any
        _protected_headers.update(self._extra_header_parameters)

        self.sd_jwt.add_signature(
            self._issuer_key,
            alg=self._sign_alg,
            protected=dumps(_protected_headers),
        )

        self.serialized_sd_jwt = self.sd_jwt.serialize(
            compact=(self._serialization_format == "compact")
        )

        # If serialization_format is "json", then add the disclosures to the JSON.
        # There does not seem to be a straightforward way to do that with the library
        # other than JSON-decoding the JWS and JSON-encoding it again.
        if self._serialization_format == "json":
            jws_content = loads(self.serialized_sd_jwt)
            jws_content[self.JWS_KEY_DISCLOSURES] = [d.b64 for d in self.ii_disclosures]
            self.serialized_sd_jwt = dumps(jws_content)

    def _create_combined(self):
        if self._serialization_format == "compact":
            self.sd_jwt_issuance = self._combine(
                self.serialized_sd_jwt, *(d.b64 for d in self.ii_disclosures)
            )
            self.sd_jwt_issuance += self.COMBINED_SERIALIZATION_FORMAT_SEPARATOR
        else:
            self.sd_jwt_issuance = self.serialized_sd_jwt



from sd_jwt.common import SDObj
import yaml
import sys


def load_yaml_specification(file):
    # create new resolver for tags
    with open(file, "r") as f:
        example = _yaml_load_specification(f)

    for property in ("user_claims", "holder_disclosed_claims"):
        if property not in example:
            sys.exit(f"Specification file must define '{property}'.")

    return example

def _yaml_load_specification(f):
    resolver = yaml.resolver.Resolver()

    # Define custom YAML tag to indicate selective disclosure
    class SDKeyTag(yaml.YAMLObject):
        yaml_tag = "!sd"

        @classmethod
        def from_yaml(cls, loader, node):
            # If this is a scalar node, it can be a string, int, float, etc.; unfortunately, since we tagged
            # it with !sd, we cannot rely on the default YAML loader to parse it into the correct data type.
            # Instead, we must manually resolve it.
            if isinstance(node, yaml.ScalarNode):
                # If the 'style' is '"', then the scalar is a string; otherwise, we must resolve it.
                if node.style == '"':
                    mp = loader.construct_yaml_str(node)
                else:
                    resolved_type = resolver.resolve(yaml.ScalarNode, node.value, (True, False))
                    if resolved_type == "tag:yaml.org,2002:str":
                        mp = loader.construct_yaml_str(node)
                    elif resolved_type == "tag:yaml.org,2002:int":
                        mp = loader.construct_yaml_int(node)
                    elif resolved_type == "tag:yaml.org,2002:float":
                        mp =  loader.construct_yaml_float(node)
                    elif resolved_type == "tag:yaml.org,2002:bool":
                        mp = loader.construct_yaml_bool(node)
                    elif resolved_type == "tag:yaml.org,2002:null":
                        mp = None
                    else:
                        raise Exception(
                            f"Unsupported scalar type for selective disclosure (!sd): {resolved_type}; node is {node}, style is {node.style}"
                        )
                return SDObj(mp)
            elif isinstance(node, yaml.MappingNode):
                return SDObj(loader.construct_mapping(node))
            elif isinstance(node, yaml.SequenceNode):
                return SDObj(loader.construct_sequence(node))
            else:
                raise Exception(
                    "Unsupported node type for selective disclosure (!sd): {}".format(
                        node
                    )
                )
            
    return yaml.load(f, Loader=yaml.FullLoader)

"""
Takes an object that has been parsed from a YAML file and removes the SDObj wrappers.
"""
def remove_sdobj_wrappers(data):
    if isinstance(data, SDObj):
        return remove_sdobj_wrappers(data.value)
    elif isinstance(data, dict):
        return {remove_sdobj_wrappers(key): remove_sdobj_wrappers(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [remove_sdobj_wrappers(value) for value in data]
    else:
        return data



import logging
import os
import random
import secrets

from base64 import urlsafe_b64decode, urlsafe_b64encode
from dataclasses import dataclass
from hashlib import sha256
from json import loads
from typing import List

DEFAULT_SIGNING_ALG = "ES256"
SD_DIGESTS_KEY = "_sd"
DIGEST_ALG_KEY = "_sd_alg"
KB_DIGEST_KEY = "_sd_hash"
SD_LIST_PREFIX = "..."

logger = logging.getLogger("sd_jwt")


@dataclass
class SDObj:
    """This class can be used to make this part of the object selective disclosable."""

    value: any

    # Make hashable
    def __hash__(self):
        return hash(self.value)


class SDJWTHasSDClaimException(Exception):
    """Exception raised when input data contains the special _sd claim reserved for SD-JWT internal data."""

    def __init__(self, error_location: any):
        super().__init__(
            f"Input data contains the special claim '{SD_DIGESTS_KEY}' reserved for SD-JWT internal data. Location: {error_location!r}"
        )


class SDJWTCommon:
    SD_JWT_HEADER = os.getenv("SD_JWT_HEADER", "example+sd-jwt") # overwriteable with extra_header_parameters = {"typ": "other-example+sd-jwt"}
    KB_JWT_TYP_HEADER = "kb+jwt"
    JWS_KEY_DISCLOSURES = "disclosures"
    JWS_KEY_KB_JWT = "kb_jwt"
    HASH_ALG = {"name": "sha-256", "fn": sha256}

    COMBINED_SERIALIZATION_FORMAT_SEPARATOR = "~"

    unsafe_randomness = False

    def __init__(self, serialization_format):
        if serialization_format not in ("compact", "json"):
            raise ValueError(f"Unknown serialization format: {serialization_format}")
        self._serialization_format = serialization_format

    def _b64hash(self, raw):
        # Calculate the SHA 256 hash and output it base64 encoded
        return self._base64url_encode(self.HASH_ALG["fn"](raw).digest())

    def _combine(self, *parts):
        return self.COMBINED_SERIALIZATION_FORMAT_SEPARATOR.join(parts)

    def _split(self, combined):
        return combined.split(self.COMBINED_SERIALIZATION_FORMAT_SEPARATOR)

    @staticmethod
    def _base64url_encode(data: bytes) -> str:
        return urlsafe_b64encode(data).decode("ascii").strip("=")

    @staticmethod
    def _base64url_decode(b64data: str) -> bytes:
        padded = f"{b64data}{'=' * divmod(len(b64data),4)[1]}"
        return urlsafe_b64decode(padded)

    def _generate_salt(self):
        if self.unsafe_randomness:
            # This is not cryptographically secure, but it is deterministic
            # and allows for repeatable output for the generation of the examples.
            logger.warning(
                "Using unsafe randomness is not suitable for production use."
            )
            return self._base64url_encode(
                bytes(random.getrandbits(8) for _ in range(16))
            )
        else:
            return self._base64url_encode(secrets.token_bytes(16))

    def _create_hash_mappings(self, disclosurses_list: List):
        # Mapping from hash of disclosure to the decoded disclosure
        self._hash_to_decoded_disclosure = {}

        # Mapping from hash of disclosure to the raw disclosure
        self._hash_to_disclosure = {}

        for disclosure in disclosurses_list:
            decoded_disclosure = loads(
                self._base64url_decode(disclosure).decode("utf-8")
            )
            _hash = self._b64hash(disclosure.encode("ascii"))
            if _hash in self._hash_to_decoded_disclosure:
                raise ValueError(
                    f"Duplicate disclosure hash {_hash} for disclosure {decoded_disclosure}"
                )

            self._hash_to_decoded_disclosure[_hash] = decoded_disclosure
            self._hash_to_disclosure[_hash] = disclosure

    def _check_for_sd_claim(self, the_object):
        # Recursively check for the presence of the _sd claim, also
        # works for arrays and nested objects.
        if isinstance(the_object, dict):
            for key, value in the_object.items():
                if key == SD_DIGESTS_KEY:
                    raise SDJWTHasSDClaimException(the_object)
                else:
                    self._check_for_sd_claim(value)
        elif isinstance(the_object, list):
            for item in the_object:
                self._check_for_sd_claim(item)
        else:
            return

    def _parse_sd_jwt(self, sd_jwt):

        if self._serialization_format == "compact":
            (
                self._unverified_input_sd_jwt,
                *self._input_disclosures,
                self._unverified_input_key_binding_jwt
            ) = self._split(sd_jwt)

            # Extract only the body from SD-JWT without verifying the signature
            _, jwt_body, _ = self._unverified_input_sd_jwt.split(".")
            self._unverified_input_sd_jwt_payload = loads(
                self._base64url_decode(jwt_body)
            )

        else:
            # if the SD-JWT is in JSON format, parse the json and extract the disclosures.
            self._unverified_input_sd_jwt = sd_jwt
            self._unverified_input_sd_jwt_parsed = loads(sd_jwt)
            self._input_disclosures = self._unverified_input_sd_jwt_parsed[
                self.JWS_KEY_DISCLOSURES
            ]
            self._unverified_input_key_binding_jwt = (
                self._unverified_input_sd_jwt_parsed.get(self.JWS_KEY_KB_JWT, "")
            )
            self._unverified_input_sd_jwt_payload = loads(
                self._base64url_decode(self._unverified_input_sd_jwt_parsed["payload"])
            )
